#include "iogateway/RawDataMessageIOGateway.h"
#include "util/StringTokenizer.h"
#include "DirectorSession.h"

namespace theater {

const char * STAGE_NAME = "Stage";

DirectorSession :: 
DirectorSession(bool instant, DataIORef scriptRef, const char * server, uint16 port) : ActorSession(STAGE_NAME), _server(server), _port(port), _instant(instant), _nextPulseTime(0), _lineNumber(0), _pace(15), _scriptEnded(false)
{
   _scriptGateway.SetDataIO(scriptRef);
}

DirectorSession ::
~DirectorSession()
{
   // empty
}

void DirectorSession :: AboutToDetachFromServer()
{
   EndServer();
   ActorSession::AboutToDetachFromServer();
}

status_t DirectorSession :: AttachedToServer()
{
   if ((ActorSession::AttachedToServer() == B_NO_ERROR)&&(_actorNameToSessionID.Put(STAGE_NAME, GetSessionIDString()) == B_NO_ERROR))
   {
      _currentSessionID = GetSessionIDString();
      return B_NO_ERROR;
   }
   return B_ERROR;
}

int32 DirectorSession :: ExecuteBufferedLine()
{
   String temp = _bufferedLine;  // store it separately so we can clear it to a known state before calling ExecuteScriptLine()
   _bufferedLine = "";
   return ExecuteScriptLine(temp, false);  // don't scan for names, since we did that when we added the buffer
}

void DirectorSession :: MessageReceivedFromStdin(MessageRef msg)
{
   _scriptQueue.AddTail(msg);
   if (_nextPulseTime == MUSCLE_TIME_NEVER) _nextPulseTime = 0;
   InvalidatePulseTime();
}

void DirectorSession :: StdinClosed()
{
   _scriptEnded = true;
   if (_nextPulseTime == MUSCLE_TIME_NEVER) _nextPulseTime = 0;
   InvalidatePulseTime();
}

void DirectorSession :: Pulse(uint64, uint64)
{
   if ((_scriptEnded)&&(_bufferedLine.Length() == 0))
   {
      AbstractMessageIOGateway * gw = GetGateway();
      if (gw->HasBytesToOutput() == false) EndServer();  // don't quit the server until all data has been sent!
   }
   else
   {
      int32 ret = -1;

      if (_bufferedLine.Length() > 0) ret = ExecuteBufferedLine();
      else
      {
         if ((_currentScriptMessage())&&(_currentScriptMessage()->HasName(PR_NAME_TEXT_LINE) == false)) _currentScriptMessage.Reset();
         if (_currentScriptMessage() == NULL) _scriptQueue.RemoveHead(_currentScriptMessage);
         if (_currentScriptMessage() == NULL)
         {
            const uint32 SCRIPT_READ_CHUNK = 10*1024;  // Read 10k worth of text into memory at once
            if (_scriptGateway.GetDataIO()) (void) _scriptGateway.DoInput(_scriptQueue, SCRIPT_READ_CHUNK);
            _scriptQueue.RemoveHead(_currentScriptMessage);
         }
         if (_currentScriptMessage())
         {
            String nextLine;
            if (_currentScriptMessage()->FindString(PR_NAME_TEXT_LINE, nextLine) == B_NO_ERROR)   
            {
               _lineNumber++;
               _currentScriptMessage()->RemoveData(PR_NAME_TEXT_LINE);
               ret = ExecuteScriptLine(nextLine.Trim(), true);
            }
         }
      }

      if ((_scriptGateway.GetDataIO() == NULL)||(ret >= 0)) _nextPulseTime = (ret >= 0) ? (GetCurrentTime64()+((uint64)(ret*1000))) : MUSCLE_TIME_NEVER;
      else
      {
         LogTime(MUSCLE_LOG_INFO, "End of script at line %lu.\n", _lineNumber);
         _nextPulseTime = MUSCLE_TIME_NEVER;
         _scriptEnded = true;  // go into wait-for-data-to-be-sent-before-quitting mode
      }
   }
}

uint64 DirectorSession :: GetPulseTime(uint64 now, uint64)
{
   return _scriptEnded ? (now + 250000) : (_instant ? now : _nextPulseTime);
}
                                                               
int32 DirectorSession :: ExecuteScriptLine(const String & l, bool scanForNames)
{
   String line = l;
   int commentStart = line.IndexOf("# ");
   if (commentStart >= 0) line = line.Substring(0, commentStart).Trim();

   if (line.Length() > 0)
   {
      String lowerLine = line.ToLowerCase();

      if (line[0] == '/')
      {
         StringTokenizer tok(line());
         String cmd = tok();
         cmd = cmd.ToLowerCase();
         if (cmd == "/enter")
         {
            String characterName = tok.GetRemainderOfString();
            characterName = characterName.Trim();
            while(characterName.EndsWith(".")) characterName = characterName.Substring(0, characterName.Length()-1);

            if (_actorNameToSessionID.ContainsKey(characterName)) LogTime(MUSCLE_LOG_ERROR, "Character [%s] is already on stage at line %lu\n", characterName(), _lineNumber);
            else
            {
               int actorSocket = Connect(_server(), _port, "Theater");
               if (actorSocket >= 0)
               {
                  AbstractReflectSessionRef actorRef(newnothrow ActorSession(characterName), NULL);
                  if ((actorRef() == NULL)||(_actorNameToSessionID.Put(characterName, actorRef()->GetSessionIDString()) != B_NO_ERROR)||(AddNewSession(actorRef, actorSocket) != B_NO_ERROR)) 
                  {
                     _actorNameToSessionID.Remove(characterName);
                     CloseSocket(actorSocket);
                  }
               }
            }
            line = "";
         }
         else if (cmd == "/exit")
         {
            String rest = tok.GetRemainderOfString();
            rest = rest.Trim();

            while(rest.EndsWith(".")) rest = rest.Substring(0, rest.Length()-1);
            StringMatcher matcher(rest.Trim());
            HashtableIterator<String, String> iter = _actorNameToSessionID.GetIterator();
            const String * nextCharacterName;
            const String * nextSessionID;
            bool foundMatch = false;
            while(((nextCharacterName = iter.GetNextKey()) != NULL)&&((nextSessionID = iter.GetNextValue()) != NULL))
            {
               if (matcher.Match(nextCharacterName->Cstr()))
               {
                  AbstractReflectSessionRef s = GetSession(nextSessionID->Cstr());
                  if ((s())&&(s() != this))   // the stage isn't allowed to walk off the stage ;^)
                  {
                     foundMatch = true;
                     s()->EndSession();
                     _actorNameToSessionID.Remove(*nextCharacterName);
                  }
               }
            }
            if (foundMatch == false) LogTime(MUSCLE_LOG_ERROR, "No characters matched [%s] to exit at line %lu\n", matcher.GetPattern()(), _lineNumber);
            line = "";
         }
         else if (cmd == "/status")
         {
            StringMatcher matcher(tok());
            String newStatus = String(tok.GetRemainderOfString()).Trim();
            HashtableIterator<String, String> iter = _actorNameToSessionID.GetIterator();
            const String * nextCharacterName;
            const String * nextSessionID;
            bool foundMatch = false;
            while(((nextCharacterName = iter.GetNextKey()) != NULL)&&((nextSessionID = iter.GetNextValue()) != NULL))
            {
               if (matcher.Match(nextCharacterName->Cstr()))
               {
                  ActorSession * s = dynamic_cast<ActorSession *>(GetSession(nextSessionID->Cstr())());
                  if (s) 
                  {
                     s->SetStatus(newStatus);
                     foundMatch = true;
                  }
               }
            }
            if (foundMatch == false) LogTime(MUSCLE_LOG_ERROR, "No characters matched [%s] to set status at line %lu\n", matcher.GetPattern()(), _lineNumber);
            line = "";
         }
         else if (cmd == "/delay")
         {
            const char * howLong = tok();
            int32 delay = howLong ? atol(howLong) : 0;
            String rest = tok.GetRemainderOfString();
            rest = rest.Trim();
            if (rest.Length() > 0) 
            {
               (void) ExecuteScriptLine(rest, true);
               (void) ExecuteBufferedLine();  // if he buffered up, execute that immediately too
            }
            return muscleMax(delay, (int32)0); 
         }
         else if (cmd == "/pace")
         {
            const char * speed = tok();
            _pace = muscleClamp((uint32)(speed ? atol(speed) : 8), (uint32)1, (uint32)1000);
            LogTime(MUSCLE_LOG_INFO, "Reading pace set to %lu chars/second\n", _pace);
            return 0;
         }
      }
      else if (scanForNames)
      {
         if (lowerLine.StartsWith("stage."))
         {
            if (_currentSessionID != GetSessionIDString()) return SetCurrentSession(GetSessionIDString(), line, 6);
                                                      else line = line.Substring(6).Trim();
         }
         else
         {
            // See if this is the start of a new actor's line, or not.
            HashtableIterator<String, String> iter = _actorNameToSessionID.GetIterator();
            const String * nextCharacterName;
            const String * nextSessionID;
            while(((nextCharacterName = iter.GetNextKey()) != NULL)&&((nextSessionID = iter.GetNextValue()) != NULL))
            {
               int32 startBytes = -1;

               // See if the character is referred to by their entire name
               {
                  String wholeName = nextCharacterName->ToLowerCase() + ".";
                  if (lowerLine.StartsWith(wholeName)) startBytes = wholeName.Length();
               }

               // If not that, see if the character is referred to by the first word in their name only....
               {
                  StringTokenizer nameTok(nextCharacterName->Cstr());
                  String firstWordInCharacterName = nameTok();  firstWordInCharacterName += ".";
                  firstWordInCharacterName = firstWordInCharacterName.ToLowerCase();
                  if (lowerLine.StartsWith(firstWordInCharacterName)) startBytes = firstWordInCharacterName.Length();
               }

               if (startBytes >= 0)
               {
                  if (_currentSessionID != *nextSessionID) return SetCurrentSession(*nextSessionID, line, startBytes);
                  else
                  {
                     line = line.Substring(startBytes).Trim();
                     break;
                  }
               }
            }
         }
      }

      if (line.Length() > 0)
      {
         ActorSession * s = dynamic_cast<ActorSession *>(GetSession(_currentSessionID())());
         if (s) s->ReciteLine(line);
           else LogTime(MUSCLE_LOG_ERROR, "No current actor to read [%s] at line %lu!\n", line(), _lineNumber);
      } 
   }
   return line.Length() * (1000 / _pace);  // read lines at _pace chars per second by default
}

int32 DirectorSession :: SetCurrentSession(const String & newSessionID, const String & line, int32 startBytes)
{
   _currentSessionID = newSessionID;
   _bufferedLine = line.Substring(startBytes).Trim();  // save this to be said after the pause
   return 1000;  // 1 second pause to switch actors
}

};  // end namespace theater
