#include <stdio.h>
#include <stdlib.h>

#include "dataio/TCPSocketDataIO.h"
#include "iogateway/MessageIOGateway.h"
#include "reflector/StorageReflectConstants.h"
#include "regex/PathMatcher.h"
#include "system/SetupSystem.h"
#include "util/MiscUtilityFunctions.h"
#include "util/NetworkUtilityFunctions.h"

using namespace muscle;

// Stolen from beshare/ShareNetClient.h
enum 
{
   NET_CLIENT_CONNECTED_TO_SERVER = 0,
   NET_CLIENT_DISCONNECTED_FROM_SERVER,
   NET_CLIENT_NEW_CHAT_TEXT,
   NET_CLIENT_CONNECT_BACK_REQUEST,
   NET_CLIENT_CHECK_FILE_COUNT,
   NET_CLIENT_PING,
   NET_CLIENT_PONG,
   NET_CLIENT_SCAN_THREAD_REPORT
};

// Also stolen from beshare/ShareNetClient.h
enum
{
   ROOT_DEPTH = 0,         // root node
   HOST_NAME_DEPTH,
   SESSION_ID_DEPTH,
   BESHARE_HOME_DEPTH,     // used to separate our stuff from other (non-BeShare) data on the same server
   USER_NAME_DEPTH,        // user's handle node would be found here
   FILE_INFO_DEPTH         // user's shared file list is here
};        

void PrintDelayCommand(uint64 & lastTime)
{
   uint64 now = GetRunTime64();
   if (now > lastTime)
   {
      uint64 millis = (now-lastTime)/1000;
      if (millis > 0) printf("/delay %llu\n", millis);
      lastTime = now;
   }
}

void DecrementName(Hashtable<String, uint32> & nameCount, Hashtable<String, String> & nameToStatus, const String & name, uint64 & prevEventTime)
{
   uint32 * count = nameCount.Get(name);
   if ((count)&&(--(*count) == 0))
   {
      PrintDelayCommand(prevEventTime);
      printf("/exit %s\n", name());
      nameCount.Remove(name); 
      nameToStatus.Remove(name);
   }
}

void IncrementName(Hashtable<String, uint32> & nameCount, Hashtable<String, String> & nameToStatus, const String & name, uint64 & prevEventTime)
{
   uint32 * count = nameCount.Get(name);
   if (count) (*count)++;
   else
   {
      nameToStatus.Remove(name);  // paranoia
      nameCount.Put(name, 1);
      PrintDelayCommand(prevEventTime);
      printf("/enter %s\n", name());
   }
}

// This program logs in to the specified server and logs the chat there
// into a MasterBesTheater-compatible script that is printed to stdout. 
// Try:  ./stenographer server=beshare.tycomsystems.com > myscript.txt
//  or:  ./stenographer server=beshare.tycomsystems.com | theaterd script=stdin testmode
// (to save a script or mirror a server's users, respectively)
int main(int argc, char ** argv)
{
   CompleteSetupSystem css;  // set up our environment

   const char * server  = "127.0.0.1";
   const char * portStr = "2960";

   Message args;
   if (ParseArgs(argc, argv, args) == B_NO_ERROR)
   {
      (void) args.FindString("server", &server);
      (void) args.FindString("port",   &portStr);
   }

   int port = atoi(portStr);
   if (port == 0) port = 2960;

   int sock = Connect(server, port);
   if (sock >= 0)
   {
      printf("# This script was recorded by the MastersBesTheater Stenographer program\n");
      printf("# From server %s:%i on %s, %s\n", server, port, __DATE__, __TIME__);
      printf("\n/pace 1000  # (Since we'll be using the /delay command to control pacing)\n");

      TCPSocketDataIO tcpIO(sock, false);
      MessageIOGateway gw;
      gw.SetDataIO(DataIORef(&tcpIO, NULL, false));

      // Upload some minimal credentials to the server so we'll get the chat texts
      {
         MessageRef cred = GetMessageFromPool();
         if (cred())
         {
            cred()->AddString("name", "StenographerBot");
            cred()->AddInt32("port", 0);
            cred()->AddBool("bot", true);
            MessageRef setMsg = GetMessageFromPool(PR_COMMAND_SETDATA);
            if ((setMsg())&&(setMsg()->AddMessage("beshare/name", cred) == B_NO_ERROR)) gw.AddOutgoingMessage(setMsg);
         }
      }

      // And our status will let people know they are being recorded
      {
         MessageRef status = GetMessageFromPool();
         if (status())
         {
            status()->AddString("userstatus", "RECORDING - BEEP!");
            MessageRef setMsg = GetMessageFromPool(PR_COMMAND_SETDATA);
            if ((setMsg())&&(setMsg()->AddMessage("beshare/userstatus", status) == B_NO_ERROR)) gw.AddOutgoingMessage(setMsg);
         }
      }

      // Subscribe to people's names and status
      {
         MessageRef subscribe = GetMessageFromPool(PR_COMMAND_SETPARAMETERS);
         if (subscribe())
         {
            subscribe()->AddBool("SUBSCRIBE:beshare/(name,userstatus)", true);
            gw.AddOutgoingMessage(subscribe);
         }
      }

      // Now we listen to incoming Messages and print out the corresponding
      // script commands, until we are control-C'd or disconnected.
      Hashtable<uint32, String> sessionIDToName;
      Hashtable<String, uint32> nameCount;
      Hashtable<String, String> nameToStatus;
      QueueGatewayMessageReceiver queueReceiver;
      uint64 prevEventTime = GetRunTime64();
      fd_set readSet, writeSet;
      while(true)
      {
         FD_ZERO(&readSet);
         FD_ZERO(&writeSet);
      
         FD_SET(sock, &readSet);
         if (gw.HasBytesToOutput()) FD_SET(sock, &writeSet);
         if (select(sock+1, &readSet, &writeSet, NULL, NULL) < 0) break;
            
         bool reading = FD_ISSET(sock, &readSet);
         bool writing = FD_ISSET(sock, &writeSet);
         bool writeError = ((writing)&&(gw.DoOutput() < 0));
         bool readError  = ((reading)&&(gw.DoInput(queueReceiver) < 0));
      
         bool hadMessages = false;
         MessageRef next;
         while(queueReceiver.RemoveHead(next) == B_NO_ERROR)
         {
            hadMessages = true;

            const Message * msg = next();
            switch(msg->what)
            {
               case PR_RESULT_DATAITEMS:
               {
                  MessageFieldNameIterator it = msg->GetFieldNameIterator(B_MESSAGE_TYPE);
                  String nodePath;
                  while(it.GetNextFieldName(nodePath) == B_NO_ERROR)
                  {
                     if (GetPathDepth(nodePath()) == USER_NAME_DEPTH)
                     {
                        String sessionID = GetPathClauseString(SESSION_ID_DEPTH, nodePath());
                        const char * nodeName = GetPathClause(USER_NAME_DEPTH, nodePath());

                        MessageRef tempRef;
                        for (int32 i=0; msg->FindMessage(nodePath, i, tempRef) == B_NO_ERROR; i++)
                        {
                           if (strncmp(nodeName, "name", 4) == 0)
                           {
                              String name;
                              if (tempRef()->FindString("name", name) == B_NO_ERROR)
                              {
                                 name.Replace('.', ' ');  // avoid possible conflict with MBT's name-ending convention

                                 int32 sid = atol(sessionID());

                                 String * oldName = sessionIDToName.Get(sid);
                                 if (oldName) DecrementName(nameCount, nameToStatus, *oldName, prevEventTime);

                                 (void) sessionIDToName.Put(sid, name);
                                 IncrementName(nameCount, nameToStatus, name, prevEventTime);
                              }
                           }
                           else if (strncmp(nodeName, "userstatus", 10) == 0)
                           {
                              String userStatus;
                              if (tempRef()->FindString("userstatus", userStatus) == B_NO_ERROR)
                              {
                                 const String * name = sessionIDToName.Get(atol(sessionID()));
                                 if (name)
                                 {
                                    const String * oldStatus = nameToStatus.Get(*name);
                                    if ((oldStatus == NULL)||(*oldStatus != userStatus))
                                    {
                                       PrintDelayCommand(prevEventTime);
                                       printf("/status %s %s\n", name->Cstr(), userStatus());
                                       nameToStatus.Put(*name, userStatus);
                                    }
                                 }
                              }
                           }
                           break;
                        }
                     }
                  }

                  // Look for sub-messages that indicate that nodes were removed from the tree
                  for (int32 i=0; (msg->FindString(PR_NAME_REMOVED_DATAITEMS, i, nodePath) == B_NO_ERROR); i++)
                  {
                     if ((GetPathDepth(nodePath()) == USER_NAME_DEPTH)&&(strncmp(GetPathClause(USER_NAME_DEPTH, nodePath()), "name", 4) == 0))
                     {
                        String sessionID = GetPathClauseString(SESSION_ID_DEPTH, nodePath());
                        String name;
                        if (sessionIDToName.Remove(atol(sessionID()), name) == B_NO_ERROR) DecrementName(nameCount, nameToStatus, name, prevEventTime);
                     }
                  }
               }
               break;
                     
               case NET_CLIENT_NEW_CHAT_TEXT:
               {
                  String text;
                  String session;
                  if ((msg->FindString("text", text) == B_NO_ERROR)&&(msg->FindString(PR_NAME_SESSION, session) == B_NO_ERROR))
                  {
                     const String * userName;
                     if ((msg->HasName("private") == false)&&((userName = sessionIDToName.Get(atol(session()))) != NULL))
                     {
                        PrintDelayCommand(prevEventTime);
                        printf("%s. %s\n", userName->Cstr(), text());
                     }
                  }
               }
               break;
            }
         }
         if (hadMessages) fflush(stdout);  // in case we get CTRL-C'd, make sure it's all actually on disk

         if ((readError)||(writeError)) break;
      }
   }
   return 0;
}
