#include <stdio.h>
#include <stdlib.h>

#include "dataio/FileDataIO.h"
#include "dataio/FileDescriptorDataIO.h"
#include "system/SetupSystem.h"
#include "reflector/FilterSessionFactory.h"
#include "util/NetworkUtilityFunctions.h"
#include "util/MiscUtilityFunctions.h"
#include "TheaterServer.h"
#include "DirectorSession.h"
#include "ActorSession.h"

using namespace theater;

// A special class that is used only when we want to read commands from stdin
class StdinSession : public StorageReflectSession
{
public:
   StdinSession(DirectorSession & director) : _director(director)
   {
      // empty
   }
  
   virtual ~StdinSession()
   {
      // empty
   }

   virtual AbstractMessageIOGateway * CreateGateway() {return newnothrow PlainTextMessageIOGateway;}
   virtual DataIO * CreateDataIO(int socket) {return newnothrow FileDescriptorDataIO(socket, false);}
   virtual void MessageReceivedFromGateway(MessageRef msgRef, void *) {_director.MessageReceivedFromStdin(msgRef);}
   virtual void MessageReceivedFromSession(AbstractReflectSession &, MessageRef, void *) {/* empty */}

   virtual bool ClientConnectionClosed()
   {
      _director.StdinClosed();
      return AbstractReflectSession::ClientConnectionClosed();
   }

   virtual const char * GetTypeName() const {return "stdin-reader";}

private:
   DirectorSession & _director;
};

int main(int argc, char ** argv)
{
   CompleteSetupSystem css;  // set up our environment

   const char * script  = NULL;
   const char * server  = "127.0.0.1";
   const char * portStr = "";
   bool testmode = false;

   Message args;
   if (ParseArgs(argc, argv, args) == B_NO_ERROR)
   {
      if (args.HasName("daemon"))
      {
         LogTime(MUSCLE_LOG_INFO, "Spawning off a theaterd daemon-child...\n");
         if (BecomeDaemonProcess(NULL, "/dev/console") != B_NO_ERROR)
         {
            LogTime(MUSCLE_LOG_CRITICALERROR, "Couldn't spawn!\n");
            return 10;
         }
      }
      (void) args.FindString("script", &script);
      (void) args.FindString("port",   &portStr);
      // Note that I deliberately don't look for a server arg here... 
      // it's too dangerous to really support that, it could cause flooding ;^)
      testmode = args.HasName("testmode");
   }

   if ((script)&&(server))
   {
      int port = atoi(portStr);
      if (port == 0) port = 2960;

      bool isStdin = (strcmp(script, "stdin") == 0);
      FILE * fp = isStdin ? NULL : fopen(script, "r");
      if ((isStdin)||(fp))
      {
         FileDataIO fio(fp);
         fp = NULL;

         int directorSocket = Connect(server, port);
         if (directorSocket >= 0)
         {
            DirectorSession director(testmode, DataIORef(isStdin ? NULL : &fio, NULL, false), server, port);
            AbstractReflectSessionRef stdinSessionRef(isStdin ? newnothrow StdinSession(director) : NULL, NULL);

            TheaterServer theater;
            if ((theater.AddNewSession(AbstractReflectSessionRef(&director, NULL, false), directorSocket) == B_NO_ERROR)&&
                ((stdinSessionRef() == NULL)||(theater.AddNewSession(stdinSessionRef, STDIN_FILENO) == B_NO_ERROR)))
            {
               status_t loopRet = theater.ServerProcessLoop();
               theater.Cleanup();
               if (loopRet == B_NO_ERROR)
               {
                  LogTime(MUSCLE_LOG_INFO, "theaterd exiting.\n");
                  return(0);
               }
               else LogTime(MUSCLE_LOG_CRITICALERROR, "theaterd exiting with an error condition.\n");
            }
            else CloseSocket(directorSocket);
         }
         else LogTime(MUSCLE_LOG_CRITICALERROR, "Couldn't connect to server [%s:%i]\n", server, port);

         if (fp) fclose(fp);
      }
      else LogTime(MUSCLE_LOG_CRITICALERROR, "Couldn't open script file [%s]\n", script);
   }
   else LogTime(MUSCLE_LOG_CRITICALERROR, "Usage: ./theaterd script=Hamlet.txt [port=2960] [testmode] [daemon]\n");

   return 10;
}
